#!/usr/bin/env python3
# (c) 2024, Cantab Research Ltd.
"""
Command-line interface
"""

import json
import logging
import ssl
import sys
from dataclasses import dataclass
from socket import gaierror
from typing import Any, Dict

import httpx
from websockets.exceptions import WebSocketException

from speechmatics_flow.cli_parser import parse_args
from speechmatics_flow.client import WebsocketClient
from speechmatics_flow.exceptions import ConversationError
from speechmatics_flow.models import (
    AudioSettings,
    ConversationConfig,
    ServerMessageType,
    Interaction,
    ConnectionSettings,
)

LOGGER = logging.getLogger(__name__)


@dataclass
class Transcripts:
    user_transcript: str = ""
    agent_transcript: str = ""


def get_log_level(verbosity):
    """
    Returns the appropriate log level given a verbosity level.

    :param verbosity: Verbosity level.
    :type verbosity: int

    :return: The logging level (eg. logging.INFO).
    :rtype: int

    :raises SystemExit: If the given verbosity level is invalid.
    """
    try:
        log_level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbosity]

        return log_level
    except KeyError as error:
        key = int(str(error))
        raise SystemExit(
            f"Only supports 2 log levels eg. -vv, you are asking for " f"-{'v' * key}"
        ) from error


def get_connection_settings(args):
    """
    Helper function which returns a ConnectionSettings object based on the
    command line options given to the program.

    :param args: Keyword arguments, typically from the command line.
    :type args: dict

    :return: Settings for the WebSocket connection.
    :rtype: models.ConnectionSettings
    """
    auth_token = args.get("auth_token")
    url = args.get("url")
    settings = ConnectionSettings(
        url=url,
        auth_token=auth_token,
        generate_temp_token=args.get("generate_temp_token") == "true",
    )

    if args.get("buffer_size") is not None:
        settings.message_buffer_size = args["buffer_size"]

    if args.get("ssl_mode") == "insecure":
        settings.ssl_context.check_hostname = False
        settings.ssl_context.verify_mode = ssl.CERT_NONE
    elif args.get("ssl_mode") == "none":
        settings.ssl_context = None

    return settings


def get_conversation_config(
    args,
):
    """
    Helper function which returns a ConversationConfig object based on the
    command line options given to the program.

    :param args: Keyword arguments probably from the command line.
    :type args: Dict

    :return: Settings for the ASR engine.
    :rtype: models.ConversationConfig
    """

    config: Dict[str, Any] = {}
    if args.get("config_file"):
        with open(args["config_file"], encoding="utf-8") as config_file:
            config = json.load(config_file)

    if config.get("conversation_config"):
        config.update(config.pop("conversation_config"))

    return ConversationConfig(**config)


def get_audio_settings(args):
    """
    Helper function which returns an AudioSettings object based on the command
    line options given to the program.

    Args:
        args (dict): Keyword arguments, typically from the command line.

    Returns:
        models.AudioSettings: Settings for the audio stream
            in the connection.
    """
    settings = AudioSettings(
        sample_rate=args.get("sample_rate"),
        chunk_size=args.get("chunk_size"),
        encoding=args.get("raw"),
    )
    return settings


# pylint: disable=too-many-arguments,too-many-statements
def add_printing_handlers(
    api,
    transcripts,
    print_json=False,
):
    """
    Adds a set of handlers to the websocket client which print out transcripts
    as they are received. This includes partials if they are enabled.

    Args:
        api (client.WebsocketClient): Client instance.
        transcripts (Transcripts): Allows the transcripts to be concatenated to
            produce a final result.
        print_json (bool, optional): Whether to print json transcript messages.
    """
    escape_seq = "\33[2K" if sys.stdout.isatty() else ""

    def convert_to_txt(message):
        if print_json:
            print(json.dumps(message))
            return
        transcript = message["metadata"]["transcript"]
        return transcript.replace("<sb> ", "").replace("</sb>", "").replace("<sb>", "")

    def transcript_handler(message):
        plaintext = convert_to_txt(message)
        if plaintext:
            sys.stdout.write(f"{escape_seq}{plaintext}\n")
        transcripts.user_transcript += plaintext

    def partial_transcript_handler(message):
        plaintext = convert_to_txt(message)
        if plaintext:
            sys.stderr.write(f"{escape_seq}{plaintext}\r")

    def prompt_handler(message):
        if print_json:
            print(json.dumps(message))
            return
        new_response = message["prompt"]["response"]
        new_plaintext_response = new_response.replace("<sb> ", "").replace("</sb> ", "")
        if new_plaintext_response:
            sys.stdout.write(f"{escape_seq}{new_plaintext_response}\n")
        transcripts.user_transcript += new_plaintext_response

    def end_of_transcript_handler(_):
        print("\n", file=sys.stderr)

    api.add_event_handler(ServerMessageType.prompt, prompt_handler)
    api.add_event_handler(ServerMessageType.AddTranscript, transcript_handler)
    api.add_event_handler(
        ServerMessageType.AddPartialTranscript, partial_transcript_handler
    )
    api.add_event_handler(ServerMessageType.EndOfTranscript, end_of_transcript_handler)


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
def main(args=None):
    """
    Main entrypoint.

    :param args: command-line arguments; defaults to None in which
            case arguments will be retrieved from `sys.argv` (this is useful
            mainly for unit tests).
    :type args: List[str]
    """
    if not args:
        args = vars(parse_args())

    logging.basicConfig(level=get_log_level(args["verbose"]))
    LOGGER.info("Args: %s", args)

    try:
        flow_main(args)
    except (KeyboardInterrupt, ValueError, ConversationError, KeyError) as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(f"{type(error).__name__}: {error}")
    except FileNotFoundError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"FileNotFoundError: {error.strerror}: '{error.filename}'."
            + " Check to make sure the filename is spelled correctly, and that the file exists."
        )
    except httpx.HTTPStatusError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(error.response.text)
    except httpx.HTTPError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(f"httpx.HTTPError: An unexpected http error occurred. {error}")
    except ConnectionResetError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"ConnectionResetError: {error}.\n\nThe most likely reason for this is that the client "
            + "has been configured to use SSL but the server does not support SSL. "
            + "If this is the case then try using --ssl-mode=none"
        )
    except (WebSocketException, gaierror) as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"WebSocketError: An unexpected error occurred in the websocket: {error}.\n\n"
            + "Check that the url and config provided is valid, "
            + "and that the language in the url matches the config.\n"
        )


def flow_main(args):
    """Main dispatch for "flow" mode commands.

    :param args: arguments from parse_args()
    :type args: argparse.Namespace
    """
    conversation_config = get_conversation_config(args)
    settings = get_connection_settings(args)
    api = WebsocketClient(settings)
    transcripts = Transcripts()
    add_printing_handlers(
        api,
        transcripts,
        print_json=args["print_json"],
    )

    def run(stream):
        try:
            api.run_synchronously(
                [Interaction(stream)],
                get_audio_settings(args),
                conversation_config,
                from_cli=True,
            )
        except KeyboardInterrupt:
            # Gracefully handle Ctrl-C, else we get a huge stack-trace.
            LOGGER.warning("Keyboard interrupt received.")

    run(sys.stdin.buffer)


if __name__ == "__main__":
    main()



# (c) 2024, Cantab Research Ltd.
"""
Helper functions used by the library.
"""

import asyncio
import concurrent.futures
import inspect
import json


def json_utf8(func):
    """A decorator to turn a function's return value into JSON"""

    def wrapper(*args, **kwargs):
        """wrapper"""
        return json.dumps(func(*args, **kwargs))

    return wrapper


async def read_in_chunks(stream, chunk_size):
    """
    Utility method for reading in and yielding chunks

    :param stream: file-like object to read audio from
    :type stream: io.IOBase

    :param chunk_size: maximum chunk size in bytes
    :type chunk_size: int

    :raises ValueError: if no data was read from the stream

    :return: a sequence of chunks of data where the length in bytes of each
        chunk is <= max_sample_size and a multiple of max_sample_size
    :rtype: collections.AsyncIterable

    """
    while True:
        # Work with both async and synchronous file readers.
        if inspect.iscoroutinefunction(stream.read):
            audio_chunk = await stream.read(chunk_size)
        else:
            # Run the read() operation in a separate thread to avoid blocking the event loop.
            with concurrent.futures.ThreadPoolExecutor() as executor:
                audio_chunk = await asyncio.get_event_loop().run_in_executor(
                    executor, stream.read, chunk_size
                )

        if not audio_chunk:
            break
        yield audio_chunk



# (c) 2024, Cantab Research Ltd.
"""
Wrapper library to interface with Flow Service API.
"""

import asyncio
import copy
import json
import logging
import os
from typing import List

import httpx
import pyaudio
import websockets

from speechmatics_flow.exceptions import (
    ConversationEndedException,
    ForceEndSession,
    ConversationError,
)
from speechmatics_flow.models import (
    ClientMessageType,
    ServerMessageType,
    AudioSettings,
    ConversationConfig,
    Interaction,
    ConnectionSettings,
)
from speechmatics_flow.utils import read_in_chunks, json_utf8

LOGGER = logging.getLogger(__name__)

# If the logging level is set to DEBUG websockets logs very verbosely,
# including a hex dump of every message being sent. Setting the websockets
# logger at INFO level specifically prevents this spam.
logging.getLogger("websockets.protocol").setLevel(logging.INFO)


class WebsocketClient:
    """
    Manage a conversation session with the agent.

    The best way to interact with this library is to instantiate this client
    and then add a set of handlers to it. Handlers respond to particular types
    of messages received from the server.

    :param connection_settings: Settings for the WebSocket connection,
        including the URL of the server.
    :type connection_settings: models.ConnectionSettings
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        connection_settings: ConnectionSettings = None,
    ):
        self.connection_settings = connection_settings
        self.websocket = None
        self.conversation_config = None
        self.audio_settings = None

        self.event_handlers = {x: [] for x in ServerMessageType}
        self.middlewares = {x: [] for x in ClientMessageType}

        self.seq_no = 0
        self.session_running = False
        self.conversation_ended_wait_timeout = 5
        self._session_needs_closing = False
        self._audio_buffer = None
        self._pyaudio = pyaudio.PyAudio

        # The following asyncio fields are fully instantiated in
        # _init_synchronization_primitives
        self._conversation_started = asyncio.Event
        self._conversation_ended = asyncio.Event
        # Semaphore used to ensure that we don't send too much audio data to
        # the server too quickly and burst any buffers downstream.
        self._buffer_semaphore = asyncio.BoundedSemaphore

    async def _init_synchronization_primitives(self):
        """
        Used to initialise synchronization primitives that require
        an event loop
        """
        self._conversation_started = asyncio.Event()
        self._conversation_ended = asyncio.Event()
        self._pyaudio = pyaudio.PyAudio()
        self._buffer_semaphore = asyncio.BoundedSemaphore(
            self.connection_settings.message_buffer_size
        )

    def _flag_conversation_started(self):
        """
        Handle a
        :py:attr:`models.ClientMessageType.ConversationStarted`
        message from the server.
        This updates an internal flag to mark the session started
        as started meaning, AddAudio is now allowed.
        """
        self._conversation_started.set()

    def _flag_conversation_ended(self):
        """
        Handle a
        :py:attr:`models.ClientMessageType.ConversationEnded`
        message from the server.
        This updates an internal flag to mark the session ended
        and server connection is closed
        """
        self._conversation_ended.set()

    @json_utf8
    def _start_conversation(self):
        """
        Constructs a
        :py:attr:`models.ClientMessageType.StartConversation`
        message.
        This initiates the conversation session.
        """
        assert self.conversation_config is not None
        msg = {
            "message": ClientMessageType.StartConversation,
            "audio_format": self.audio_settings.asdict(),
            "conversation_config": self.conversation_config.asdict(),
        }
        self.session_running = True
        self._call_middleware(ClientMessageType.StartConversation, msg, False)
        LOGGER.debug(msg)
        return msg

    @json_utf8
    def _end_of_audio(self):
        """
        Constructs an
        :py:attr:`models.ClientMessageType.AudioEnded`
        message.
        """
        msg = {"message": ClientMessageType.AudioEnded, "last_seq_no": self.seq_no}
        self._call_middleware(ClientMessageType.AudioEnded, msg, False)
        LOGGER.debug(msg)
        return msg

    async def _wait_for_conversation_ended(self):
        """
        Waits for :py:attr:`models.ClientMessageType.ConversationEnded`
        message from the server.
        """
        await asyncio.wait_for(
            self._conversation_ended.wait(), self.conversation_ended_wait_timeout
        )

    async def _consumer(self, message, from_cli: False):
        """
        Consumes messages and acts on them.

        :param message: Message received from the server.
        :type message: str

        :raises ConversationError on an error message received from the
            server after the Session started.
        :raises ConversationEndedException: on models.ServerMessageType.ConversationEnded message
            received from the server.
        :raises ForceEndSession: If this was raised by the user's event
            handler.
        """
        LOGGER.debug(message)
        if isinstance(message, (bytes, bytearray)):
            # add an audio message to local buffer only when running from cli
            if from_cli:
                await self._audio_buffer.put(message)
            # Flow service does not send message_type with binary data,
            # so we need to set it here for event_handler to work
            message_type = ServerMessageType.audio
        else:
            message = json.loads(message)
            message_type = message.get("message")

        if message_type is None:
            return

        for handler in self.event_handlers[message_type]:
            try:
                handler(copy.deepcopy(message))
            except ForceEndSession:
                LOGGER.warning("Session was ended forcefully by an event handler")
                raise

        if message_type == ServerMessageType.ConversationStarted:
            self._flag_conversation_started()
        elif message_type == ServerMessageType.AudioAdded:
            self._buffer_semaphore.release()
        elif message_type == ServerMessageType.ConversationEnded:
            self._flag_conversation_ended()
            raise ConversationEndedException()
        elif message_type == ServerMessageType.Warning:
            LOGGER.warning(message["reason"])
        elif message_type == ServerMessageType.Error:
            raise ConversationError(message["reason"])

    async def _read_from_microphone(self):
        print(
            f"Default input device: {self._pyaudio.get_default_input_device_info()['name']}"
        )
        print(
            f"Default output device: {self._pyaudio.get_default_output_device_info()['name']}"
        )
        print("Start speaking...")
        stream = self._pyaudio.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self.audio_settings.sample_rate,
            input=True,
        )
        try:
            while True:
                if self._session_needs_closing or self._conversation_ended.is_set():
                    break

                await asyncio.wait_for(
                    self._buffer_semaphore.acquire(),
                    timeout=self.connection_settings.semaphore_timeout_seconds,
                )

                # audio_chunk size is 128 * 2 = 256 bytes which is about 8ms
                audio_chunk = stream.read(num_frames=128, exception_on_overflow=False)

                self.seq_no += 1
                self._call_middleware(ClientMessageType.AddAudio, audio_chunk, True)
                await self.websocket.send(audio_chunk)
        except KeyboardInterrupt:
            await self.websocket.send(self._end_of_audio())
        finally:
            await self._wait_for_conversation_ended()
            stream.stop_stream()
            stream.close()
            self._pyaudio.terminate()

    async def _consumer_handler(self, from_cli: False):
        """
        Controls the consumer loop for handling messages from the server.

        raises: ConnectionClosedError when the upstream closes unexpectedly
        """
        while self.session_running:
            try:
                message = await self.websocket.recv()
            except websockets.exceptions.ConnectionClosedOK:
                # Can occur if a timeout has closed the connection.
                LOGGER.info("Cannot receive from closed websocket.")
                return
            except websockets.exceptions.ConnectionClosedError as ex:
                LOGGER.info("Disconnected while waiting for recv().")
                raise ex
            await self._consumer(message, from_cli)

    async def _stream_producer(self, stream, audio_chunk_size):
        async for audio_chunk in read_in_chunks(stream, audio_chunk_size):
            if self._session_needs_closing or self._conversation_ended.is_set():
                break
            await asyncio.wait_for(
                self._buffer_semaphore.acquire(),
                timeout=self.connection_settings.semaphore_timeout_seconds,
            )

            self.seq_no += 1
            self._call_middleware(ClientMessageType.AddAudio, audio_chunk, True)
            yield audio_chunk

    async def _producer_handler(self, interactions: List[Interaction]):
        """
        Controls the producer loop for sending messages to the server.
        """
        await self._conversation_started.wait()
        if interactions[0].stream.name == "<stdin>":
            return await self._read_from_microphone()

        for interaction in interactions:
            async for message in self._stream_producer(
                interaction.stream, self.audio_settings.chunk_size
            ):
                try:
                    await self.websocket.send(message)
                except Exception as e:
                    LOGGER.error(f"error sending message: {e}")
                    return
            if interaction.callback:
                interaction.callback(self)

        await self.websocket.send(self._end_of_audio())
        await self._wait_for_conversation_ended()

    async def _playback_handler(self):
        """
        Reads audio binary messages from the playback buffer and plays them to the user.
        """
        stream = self._pyaudio.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self.audio_settings.sample_rate,
            output=True,
        )
        try:
            while True:
                if self._session_needs_closing or self._conversation_ended.is_set():
                    break
                try:
                    audio_message = await self._audio_buffer.get()
                    stream.write(audio_message)
                    self._audio_buffer.task_done()
                except Exception as e:
                    LOGGER.error(f"Error during audio playback: {e}")
                    raise e
        finally:
            stream.close()
            stream.stop_stream()
            self._pyaudio.terminate()

    def _call_middleware(self, event_name, *args):
        """
        Call the middlewares attached to the client for the given event name.

        :raises ForceEndSession: If this was raised by the user's middleware.
        """
        for middleware in self.middlewares[event_name]:
            try:
                middleware(*args)
            except ForceEndSession:
                LOGGER.warning("Session was ended forcefully by a middleware")
                raise

    def add_event_handler(self, event_name, event_handler):
        """
        Add an event handler (callback function) to handle an incoming
        message from the server. Event handlers are passed a copy of the
        incoming message from the server. If `event_name` is set to 'all' then
        the handler will be added for every event.

        For example, a simple handler that just LOGGER.debugs out the
        :py:attr:`models.ServerMessageType.audio`
        messages received:

        >>> client = WebsocketClient(
                ConnectionSettings(url="wss://localhost:9000"))
        >>> handler = lambda msg: LOGGER.debug(msg)
        >>> client.add_event_handler(ServerMessageType.audio, handler)

        :param event_name: The name of the message for which a handler is
                being added. Refer to
                :py:class:`models.ServerMessageType` for a list
                of the possible message types.
        :type event_name: str

        :param event_handler: A function to be called when a message of the
            given type is received.
        :type event_handler: Callable[[dict], None]

        :raises ValueError: If the given event name is not valid.
        """
        if event_name == "all":
            for name in self.event_handlers.keys():
                self.event_handlers[name].append(event_handler)
        elif event_name not in self.event_handlers:
            raise ValueError(
                f"Unknown event name: {event_name!r}, expected to be "
                f"'all' or one of {list(self.event_handlers.keys())}."
            )
        else:
            self.event_handlers[event_name].append(event_handler)

    def add_middleware(self, event_name, middleware):
        """
        Add middleware to handle outgoing messages sent to the server.
        Middlewares are passed a reference to the outgoing message, which
        they may alter.
        If `event_name` is set to 'all' then the handler will be added for
        every event.

        :param event_name: The name of the message for which middleware is
            being added. Refer to the V2 API docs for a list of the possible
            message types.
        :type event_name: str

        :param middleware: A function to be called to process an outgoing
            message of the given type. The function receives the message as
            the first argument and a second, boolean argument indicating
            whether the message is binary data (which implies it is an
            AddAudio message).
        :type middleware: Callable[[dict, bool], None]

        :raises ValueError: If the given event name is not valid.
        """
        if event_name == "all":
            for name in self.middlewares.keys():
                self.middlewares[name].append(middleware)
        elif event_name not in self.middlewares:
            raise ValueError(
                (
                    f"Unknown event name: {event_name}, expected to be 'all'"
                    f"or one of {list(self.middlewares.keys())}."
                )
            )
        else:
            self.middlewares[event_name].append(middleware)

    async def _communicate(self, interactions: List[Interaction], from_cli=False):
        """
        Create a producer/consumer for transcription messages and
        communicate with the server.
        Internal method called from _run.
        """
        try:
            start_conversation_msg = self._start_conversation()
        except ForceEndSession:
            return
        await self.websocket.send(start_conversation_msg)

        tasks = [
            asyncio.create_task(self._consumer_handler(from_cli)),
            asyncio.create_task(self._producer_handler(interactions)),
        ]

        # Run the playback task that plays audio messages to the user when started from cli
        if from_cli:
            self._audio_buffer = asyncio.Queue()
            tasks.append(asyncio.create_task(self._playback_handler()))

        (done, pending) = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_EXCEPTION,
        )

        # If a task is pending, the other one threw an exception, so tidy up
        for task in pending:
            task.cancel()

        for task in done:
            exc = task.exception()
            if exc and not isinstance(
                exc,
                (ForceEndSession, ConversationEndedException),
            ):
                raise exc

    async def run(
        self,
        interactions: List[Interaction],
        audio_settings: AudioSettings = AudioSettings(),
        conversation_config: ConversationConfig = None,
        from_cli: bool = False,
    ):
        """
        Begin a new recognition session.
        This will run asynchronously. Most callers may prefer to use
        :py:meth:`run_synchronously` which will block until the session is
        finished.

        :param interactions: A list of interactions with FlowService API.
        :type interactions: List[Interaction]

        :param audio_settings: Configuration for the audio stream.
        :type audio_settings: models.AudioSettings

        :param conversation_config: Configuration for the conversation.
        :type conversation_config: models.ConversationConfig

        :raises Exception: Can raise any exception returned by the
            consumer/producer tasks.
        """
        self.seq_no = 0
        self.conversation_config = conversation_config
        self.audio_settings = audio_settings

        await self._init_synchronization_primitives()

        extra_headers = {}
        auth_token = self.connection_settings.auth_token
        if auth_token and self.connection_settings.generate_temp_token:
            auth_token = await get_temp_token(auth_token)
        extra_headers["Authorization"] = f"Bearer {auth_token}"
        try:
            async with websockets.connect(  # pylint: disable=no-member
                self.connection_settings.url,
                ssl=self.connection_settings.ssl_context,
                ping_timeout=self.connection_settings.ping_timeout_seconds,
                # Don't limit the max. size of incoming messages
                max_size=None,
                extra_headers=extra_headers,
            ) as self.websocket:
                await self._communicate(interactions, from_cli)
        finally:
            self.session_running = False
            self._session_needs_closing = False
            self.websocket = None

    def stop(self):
        """
        Indicates that the recognition session should be forcefully stopped.
        Only used in conjunction with `run`.
        You probably don't need to call this if you're running the client via
        :py:meth:`run_synchronously`.
        """
        self._session_needs_closing = True

    def run_synchronously(self, *args, timeout=None, **kwargs):
        """
        Run the transcription synchronously.
        :raises asyncio.TimeoutError: If the given timeout is exceeded.
        """
        # pylint: disable=no-value-for-parameter
        asyncio.run(asyncio.wait_for(self.run(*args, **kwargs), timeout=timeout))


async def get_temp_token(api_key):
    """
    Used to get a temporary token from management platform api
    """
    mp_api_url = os.getenv("SM_MANAGEMENT_PLATFORM_URL", "https://mp.speechmatics.com")
    endpoint = f"{mp_api_url}/v1/api_keys?type=flow"
    body = {"ttl": 300, "client_ref": "speechmatics-flow-python-client"}
    headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
    response = httpx.post(endpoint, json=body, headers=headers)
    response.raise_for_status()
    data = response.json()
    return data["key_value"]



# (c) 2024, Cantab Research Ltd.
"""
Exceptions and errors used by the library.
"""


class ConversationError(Exception):
    """
    Indicates an error in flow conversation session.
    """


class ForceEndSession(Exception):
    """
    Can be raised by the user from a middleware or event handler
     to force the transcription session to end early.
    """


class ConversationEndedException(Exception):
    """
    Indicates the conversation session ended.
    """



# (c) 2024, Cantab Research Ltd.
"""
Parsers used by the CLI to handle CLI arguments
"""
import argparse
import logging

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def get_arg_parser():
    """
    Creates a command-line argument parser objct

    :return: The argparser object with all commands and subcommands.
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description="CLI for Speechmatics Flow API.")
    parser.add_argument(
        "-v",
        dest="verbose",
        action="count",
        default=0,
        help=(
            "Set the log level for verbose logs. "
            "The number of flags indicate the level, eg. "
            "-v is INFO and -vv is DEBUG."
        ),
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help=(
            "Prints useful symbols to represent the messages on the wire. "
            "Symbols are printed to STDERR, use only when STDOUT is "
            "redirected to a file."
        ),
    )
    parser.add_argument(
        "--url",
        type=str,
        help="Websocket for Flow API URL (e.g. wss://flow.api.speechmatics.com/v1/flow",
    )
    parser.add_argument(
        "--auth-token",
        type=str,
        help="Authentication token to authorize the client.",
    )
    parser.add_argument(
        "--generate-temp-token",
        default="true",
        choices=["true", "false"],
        help="Automatically generate a temporary token for authentication.",
    )
    parser.add_argument(
        "--ssl-mode",
        default="regular",
        choices=["regular", "insecure", "none"],
        help=(
            "Use a preset configuration for the SSL context. With `regular` "
            "mode a valid certificate is expected. With `insecure` mode"
            " a self signed certificate is allowed."
            " With `none` then SSL is not used."
        ),
    )
    parser.add_argument(
        "--raw",
        metavar="ENCODING",
        type=str,
        default="pcm_s16le",
        help=(
            "Indicate that the input audio is raw, provide the encoding"
            "of this raw audio, eg. pcm_f32le."
        ),
    )
    parser.add_argument(
        "--sample-rate",
        type=int,
        default=16_000,
        help="The sample rate in Hz of the input audio, if in raw format.",
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        default=256,
        help=(
            "How much audio data, in bytes, to send to the server in each "
            "websocket message. Larger values can increase latency, but "
            "values which are too small create unnecessary overhead."
        ),
    )
    parser.add_argument(
        "--buffer-size",
        default=512,
        type=int,
        help=(
            "Maximum number of messages to send before waiting for"
            "acknowledgements from the server."
        ),
    )
    parser.add_argument(
        "--print-json",
        default=False,
        action="store_true",
        help=(
            "Print the JSON partial & final transcripts received rather than "
            "plaintext messages."
        ),
    )

    return parser


def parse_args(args=None):
    """
    Parses command-line arguments.

    :param args: List of arguments to parse.
    :type args: (List[str], optional)

    :return: The set of arguments provided along with their values.
    :rtype: Namespace
    """
    parser = get_arg_parser()
    parsed_args = parser.parse_args(args=args)
    return parsed_args



# (c) 2024, Cantab Research Ltd.
"""
Data models and message types used by the library.
"""

import io
import ssl
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Callable, Dict, Optional, Literal


@dataclass
class AudioSettings:
    """Defines audio parameters."""

    encoding: str = "pcm_s16le"
    """Encoding format when raw audio is used. Allowed values are
    `pcm_f32le` and `pcm_s16le`"""

    sample_rate: int = 16000
    """Sampling rate in hertz."""

    chunk_size: int = 256
    """Chunk size in bytes."""

    def asdict(self):
        return {
            "type": "raw",
            "encoding": self.encoding,
            "sample_rate": self.sample_rate,
        }


@dataclass
class ConnectionSettings:
    """Defines connection parameters."""

    url: str
    """Websocket server endpoint."""

    message_buffer_size: int = 512
    """Message buffer size in bytes."""

    ssl_context: ssl.SSLContext = field(default_factory=ssl.create_default_context)
    """SSL context."""

    semaphore_timeout_seconds: float = 120
    """Semaphore timeout in seconds."""

    ping_timeout_seconds: float = 60
    """Ping-pong timeout in seconds."""

    auth_token: Optional[str] = None
    """auth token to authenticate a customer."""

    generate_temp_token: Optional[bool] = True
    """Automatically generate a temporary token for authentication."""


@dataclass
class ConversationConfig:
    """Defines configuration parameters for conversation requests."""

    template_id: Literal[
        "default", "flow-service-assistant-amelia", "flow-service-assistant-humphrey"
    ] = "default"
    """Name of a predefined template."""

    template_variables: Optional[Dict[str, str]] = None
    """Optional parameter to allow overriding the default values of variables defined in the template."""

    def asdict(self):
        return asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )


class ClientMessageType(str, Enum):
    # pylint: disable=invalid-name
    """Defines various messages sent from client to server."""

    StartConversation = "StartConversation"
    """Initiates a conversation job based on configuration set previously."""

    AddAudio = "AddAudio"
    """Adds more audio data to the recognition job. The server confirms
    receipt by sending an :py:attr:`ServerMessageType.AudioAdded` message."""

    AudioEnded = "AudioEnded"
    """Indicates audio input has finished."""


class ServerMessageType(str, Enum):
    # pylint: disable=invalid-name
    """Defines various message types sent from server to client."""

    ConversationStarted = "ConversationStarted"
    """Server response to :py:attr:`ClientMessageType.StartConversation`,
    acknowledging that a conversation session has started."""

    AddPartialTranscript = "AddPartialTranscript"
    """Indicates a partial transcript, which is an incomplete transcript that
    is immediately produced and may change as more context becomes available.
    """

    AudioAdded = "AudioAdded"
    """Server response to :py:attr:`ClientMessageType.AddAudio`, indicating
    that audio has been added successfully."""

    AddTranscript = "AddTranscript"
    """Indicates the final transcript of a part of the audio."""

    audio = "audio"
    """Message contains binary data"""

    prompt = "prompt"
    """Message contains text data"""

    ConversationEnded = "ConversationEnded"
    """Message indicates the session ended."""

    EndOfTranscript = "EndOfTranscript"
    """Server response to :py:attr:`ClientMessageType.EndOfStream`,
    after the server has finished sending all :py:attr:`AddTranscript`
    messages."""

    Info = "Info"
    """Indicates a generic info message."""

    Warning = "Warning"
    """Indicates a generic warning message."""

    Error = "Error"
    """Indicates n generic error message."""


@dataclass
class Interaction:
    """Defines various interactions between client and server."""

    stream: io.BufferedReader
    callback: Optional[Callable] = None



#!/usr/bin/env python3
"""
Setuptools configuration for the Speechmatics Flow client
"""

import logging
import os

from setuptools import setup, find_packages


def read(fname):
    """
    Load content of the file with path relative to where setup.py is.

    Args:
        fname (str): file name (or path relative to the project root)

    Returns:
        str: file content
    """
    fpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    with open(fpath, encoding="utf-8") as path:
        return path.read()


def read_list(fname):
    """
    Load the content of the file and split it into a list of lines.

    Args:
        fname (str): file name (or path relative to the project root)

    Returns:
        List[str]: file content (one string per line) with end-of-lines
                   characters stripped off and empty lines filtered out
    """
    content = read(fname)
    retval = list(filter(None, content.split("\n")))

    return retval


def get_version(fname):
    """
    Retrieve the version from the VERSION file.

    Args:
        fname (str): file containing only the version

    Returns:
        str: version with whitespace characters stripped off
    """
    return read(fname).strip()


logging.basicConfig(level=logging.INFO)
setup(
    name="speechmatics-flow",
    version=os.getenv("VERSION", get_version("VERSION")),
    packages=find_packages(exclude=["tests"]),
    url="https://github.com/speechmatics/speechmatics-flow/",
    license="MIT",
    author="Speechmatics",
    author_email="support@speechmatics.com",
    description="Python library and CLI for Speechmatics Flow API",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    install_requires=read_list("requirements.txt"),
    tests_require=read_list("requirements-dev.txt"),
    entry_points={
        "console_scripts": ["speechmatics-flow = speechmatics_flow.cli:main"]
    },
    project_urls={
        "Documentation": "https://speechmatics.github.io/speechmatics-flow/",
        "Source Code": "https://github.com/speechmatics/speechmatics-flow/",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Topic :: Multimedia :: Sound/Audio :: Speech",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    include_package_data=True,
    python_requires=">=3.7",
)



import pytest

from speechmatics_flow import models

TEMPLATE_VARS = {
    "persona": "You are an aging English butler named Humphrey.",
    "style": "Be charming but unpredictable.",
    "context": "You are taking a customer's order for fast food.",
}


@pytest.mark.parametrize(
    "config, want",
    [
        ({}, {"type": "raw", "encoding": "pcm_s16le", "sample_rate": 16000}),
        (
            {"encoding": "pcm_f32le", "sample_rate": 44100},
            {"type": "raw", "encoding": "pcm_f32le", "sample_rate": 44100},
        ),
    ],
)
def test_audio_settings(config, want):
    audio_settings = models.AudioSettings(**config)
    got = audio_settings.asdict()
    assert got == want


@pytest.mark.parametrize(
    "config, want",
    [
        ({}, {"template_id": "default"}),
        (
            {"template_id": "test", "template_variables": TEMPLATE_VARS},
            {"template_id": "test", "template_variables": TEMPLATE_VARS},
        ),
    ],
)
def test_conversation_config(config, want):
    conversation_config = models.ConversationConfig(**config)
    got = conversation_config.asdict()
    assert got == want
